﻿using System.Collections.Generic;
using System.Linq;
using System.Reflection;
using Harmony;
using Verse;

namespace XenobionicPatcher {
    public class DefInjectors {
        public void InjectSurgeryRecipes (IEnumerable<RecipeDef> surgeryList, IEnumerable<ThingDef> pawnList) {
            Base XP = Base.Instance;

            /* Some mods (like *cough*elderthings*cough*) like to use different def names for basic body parts.
             * This makes it harder to add the surgery recipe to the alien.  We'll need to add in the similar
             * body part to the appliedOnFixedBodyParts list first.
             * 
             * First, look through the list of common surgery recipes to infer the part type.  In other words, if
             * there's a surgery called "Install bionic arm" then that part is an arm that can accept other kinds of
             * arms.  Then, also look for body part matches by looking at the body part labels directly (basically
             * duck typing).
             */
            var partToPartMapper = new Dictionary<string, List<BodyPartDef>> {};

            foreach (RecipeDef surgery in surgeryList.Where(s => s.targetsBodyPart)) {
                foreach (ThingDef pawnDef in pawnList) {
                    // We can't cross the animal/humanlike boundary with these checks because animal surgery recipes tend to be a lot
                    // looser with limbs (ie: power claws on animal legs)
                    if (XP.GetPawnBioType(pawnDef) != XP.GetSurgeryBioType(surgery)) continue;

                    // Look for matching surgery labels, and map them to similar body parts
                    foreach (RecipeDef otherSurgery in pawnDef.recipes?.Where(s =>
                        s.targetsBodyPart && s != surgery && s.defName != surgery.defName && s.label.ToLower() == surgery.label.ToLower()
                    )) {
                        foreach (BodyPartDef surgeryBodyPart in surgery.appliedOnFixedBodyParts) {
                            foreach (BodyPartDef otherSurgeryBodyPart in otherSurgery.appliedOnFixedBodyParts.Where(
                                bp => bp != surgeryBodyPart && bp.defName != surgeryBodyPart.defName && bp.label.ToLower() != surgeryBodyPart.label.ToLower()
                            )) {
                                if (!partToPartMapper.ContainsKey(surgeryBodyPart.defName)) partToPartMapper[surgeryBodyPart.defName] = new List<BodyPartDef> {};
                                partToPartMapper[surgeryBodyPart.defName].AddDistinct(otherSurgeryBodyPart);
                            }
                        }
                    }

                    // Looks for matching body part labels
                    foreach (BodyPartDef surgeryBodyPart in surgery.appliedOnFixedBodyParts) {
                        foreach (
                            BodyPartDef raceBodyPart in
                            pawnDef.race.body.AllParts.Select(bpr => bpr.def).Where(rbp =>
                                surgeryBodyPart != rbp && surgeryBodyPart.defName != rbp.defName && surgeryBodyPart.label.ToLower() == rbp.label.ToLower()
                            )
                        ) {
                            if (!partToPartMapper.ContainsKey(surgeryBodyPart.defName)) partToPartMapper[surgeryBodyPart.defName] = new List<BodyPartDef> {};
                            partToPartMapper[surgeryBodyPart.defName].AddDistinct(raceBodyPart);
                        }
                    }
                }

                // DEBUG
                /*
                if (surgery.label.ToLower() == "install power claw" || surgery.label.ToLower() == "install bionic leg") {
                    XP.ModLogger.Message(
                        "    [" + surgery.defName + "] Before: " + String.Join(", ", surgery.appliedOnFixedBodyParts?.Select(bp => bp.defName).ToArray())
                    );
                }
                */
            }

            // With the parts mapped, add new body parts to existing recipes
            foreach (RecipeDef surgery in surgeryList.Where(s => s.targetsBodyPart)) {
                var newPartList = new List<BodyPartDef> {};
                foreach (BodyPartDef surgeryBodyPart in surgery.appliedOnFixedBodyParts) {
                    if (partToPartMapper.ContainsKey(surgeryBodyPart.defName)) {
                        newPartList.AddRange(partToPartMapper[surgeryBodyPart.defName]);
                        newPartList.RemoveDuplicates();
                    }
                }
                if (newPartList.Count() >= 1) {
                    surgery.appliedOnFixedBodyParts.AddRange(newPartList);
                    surgery.appliedOnFixedBodyParts.RemoveDuplicates();
                }

                // DEBUG
                /*
                if (surgery.label.ToLower() == "install power claw" || surgery.label.ToLower() == "install bionic leg") {
                    XP.ModLogger.Message(
                        "    [" + surgery.defName + "] After: " + String.Join(", ", surgery.appliedOnFixedBodyParts?.Select(bp => bp.defName).ToArray())
                    );
                }
                */
            }

            // Apply relevant missing surgery options to all pawn Defs
            foreach (RecipeDef surgery in surgeryList) {
                foreach (ThingDef pawnDef in pawnList) {
                    bool shouldAddSurgery = false;

                    // If the pawn never had any recipes, then it doesn't even have the basics, so don't risk adding new ones
                    if (pawnDef.recipes == null || pawnDef.recipes.Count() < 1) continue;

                    // If it already exists, don't add it
                    else if (pawnDef.recipes.Any( s =>
                        s == surgery || s.defName == surgery.defName || s.label.ToLower() == surgery.label.ToLower()
                    )) continue;

                    // If it's an administer recipe, add it
                    else if (!surgery.targetsBodyPart) shouldAddSurgery = true;

                    // If it targets any body parts that exist within the alien, add it
                    else if (surgery.targetsBodyPart && surgery.appliedOnFixedBodyParts.Count() >= 1 && surgery.appliedOnFixedBodyParts.Any( sbp =>
                        pawnDef.race.body.AllParts.Any( rbpr => sbp.defName == rbpr.def.defName )
                    )) shouldAddSurgery = true;

                    if (shouldAddSurgery) {
                        pawnDef.recipes.Add(surgery);
                        if (surgery.recipeUsers == null) surgery.recipeUsers = new List<ThingDef> { pawnDef };
                        else                             surgery.recipeUsers.Add(pawnDef);
                    }
                }
            }
        }

        public void CleanupSurgeryRecipes (IEnumerable<RecipeDef> surgeryList, IEnumerable<ThingDef> pawnList) {
            Base XP = Base.Instance;

            // Just in case we have some easy dupes to clean
            foreach (ThingDef pawnDef in pawnList) {
                pawnDef.recipes?.RemoveDuplicates();
            }

            // XXX: Given that Core and other "important" modules are loaded first, we'll assume the first recipe
            // (surgery) is the destination that should receive the expanded parts lists from otherSurgery.

            // Merge surgeries that do the same thing on different body parts
            var partsSurgeryList = surgeryList.Where(s => s.targetsBodyPart).ToList();
            for (int ps = 0; ps < partsSurgeryList.Count(); ps++) {
                RecipeDef surgery = partsSurgeryList[ps];
                
                // The other side of the "easy dupe" cleaning
                surgery.recipeUsers?.RemoveDuplicates();

                var toDelete = new List<RecipeDef> {};
                foreach (RecipeDef otherSurgery in partsSurgeryList.Where(s => 
                    s != surgery && s.defName != surgery.defName &&
                    s.label.ToLower()    == surgery.label.ToLower()    &&
                    s.workerClass        == surgery.workerClass        &&
                    s.workerCounterClass == surgery.workerCounterClass &&
                    s.addsHediff         == surgery.addsHediff         &&

                    s.fixedIngredientFilter?.Summary == surgery.fixedIngredientFilter?.Summary
                )) {
                    /*
                    XP.ModLogger.Message(
                        "    [" + surgery.defName + "] Merging " + otherSurgery.defName
                    );
                    */

                    surgery.appliedOnFixedBodyParts.AddRange(otherSurgery.appliedOnFixedBodyParts);
                    surgery.appliedOnFixedBodyParts.RemoveDuplicates();

                    surgery.recipeUsers.AddRange(otherSurgery.recipeUsers);
                    surgery.recipeUsers.RemoveDuplicates();

                    // This is like pawnDef.AllRecipes, without actually initializing the permanent cache.  We aren't going
                    // to trust that every def is actually injected in both sides (pawn.recipes + surgery.recipeUsers).
                    List<ThingDef> otherSurgeryPawns = otherSurgery.recipeUsers;
                    otherSurgeryPawns.AddRange( pawnList.Where(p => p.recipes.Contains(otherSurgery)) );
                    otherSurgeryPawns.RemoveDuplicates();
                    
                    foreach (ThingDef pawnDef in otherSurgeryPawns) {
                        surgery.recipeUsers.AddDistinct(pawnDef);
                        
                        // Try to keep the same index in the replacement
                        var recipes = pawnDef.recipes;
                        int i = recipes.IndexOf(otherSurgery);
                        if (i != -1) {
                            recipes[i] = surgery;
                        }
                        else {
                            // XXX: How would we even get here???
                            recipes.Add(surgery);
                            recipes.Remove(otherSurgery);
                        }
                    }

                    // XXX: Well, of course it's a private method.  Guess we reflect now...

                    // Time to die!
                    MethodInfo removeMethod = AccessTools.Method(typeof(DefDatabase<RecipeDef>), "Remove");
                    removeMethod.Invoke(null, new object[] { otherSurgery });  // static method: first arg is null

                    toDelete.Add(otherSurgery);  // don't re-merge in the other direction on our main loop

                    /*
                    XP.ModLogger.Message(
                        "    [" + surgery.defName + "] Results: " + string.Join(", ", surgery.appliedOnFixedBodyParts?.Select(bp => bp.defName).ToArray())
                    );
                    */
                }

                // Second loop is still an enumerator, so delete here
                toDelete.ForEach( s => partsSurgeryList.Remove(s) );
            }
        }
    }
}
